/**
 *******************************************************************************
 * @file    usbd_scsi.c
 * @brief   MSC(SCSI) Function Code for TOSHIBA 'TMPM366' Device Series
 * @version V0.202
 * @date    2014/05/20
 *
 * DO NOT USE THIS SOFTWARE WITHOUT THE SOFTWARE LICENSE AGREEMENT.
 *
 * (C)Copyright TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION 2017 All rights reserved
 *******************************************************************************
 */

#include <string.h>
#include "usbd_hw.h"
#include "usbd_device_request.h"
#include "usbd_var.h"
#include "usbd_msc.h"
#include "usbd_scsi.h"
#include "usbd_trans.h"
#include "usbd_descriptor_msc.h"

/**
  * @brief Callback information for MSC(SCSI)
  */
SCSI_CB_t gSCSI_CB;
/**
  * @brief Callback information for MSC(BOT)
  */
MSC_CB_t gBOT_CB;

/**
  * @brief Don't care status
  */
bool gbStatDCSCSI;

/**
  * @brief Data length
  */
static uint32_t l_ulDataLength;
/**
  * @brief Media condition flag
  */
static uint8_t l_ucMediaCondition;
/**
  * @brief Media state
  */
static uint8_t l_ucMediaState;
/**
  * @brief Command status (Used by CSW)
  */
static uint8_t l_ucCSWStatus;
/**
  * @brief Operation code of SCSI command
  */
static uint8_t l_ucOperationCode;
/**
  * @brief Sense key (Used by reply data of Request Sense(03h) command)
  */
static uint8_t l_ucSenseKey;
/**
  * @brief Prevent bit(Used by Prevent/Allow Medium Removal(1Eh) command)
  */
static uint8_t l_ucPrevent;
/**
  * @brief Sense data for 'Error' (Used by reply data of Request Sense(03h) command)
  */
static USBD_MSC_SCSI_RET_SENSE_DATA_t l_tSense_Data_Error = {
    0x70U,
    0x00U,
    0x05U,  /* Illegal Request */
    0x00U, 0x00U, 0x00U, 0x00U,
    0x0AU,
    0x00U, 0x00U, 0x00U, 0x00U,
    0x24U,
    0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U
};
/**
  * @brief Sense data for 'No Media' (Used by reply data of Request Sense(03h) command)
  */
static USBD_MSC_SCSI_RET_SENSE_DATA_t l_tSense_Data_NoMedia = {
    0x70U,
    0x00U,
    0x02U,  /* Not Ready */
    0x00U, 0x00U, 0x00U, 0x00U,
    0x0AU,
    0x00U, 0x00U, 0x00U, 0x00U,
    0x3AU,
    0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U
};

/**
  * @brief Sense data for 'Media Change' (Used by reply data of Request Sense(03h) command)
  */
static USBD_MSC_SCSI_RET_SENSE_DATA_t l_tSense_Data_MediaChenge = {
    0x70U,
    0x00U,
    0x06U,  /* Unit Attention */
    0x00U, 0x00U, 0x00U, 0x00U,
    0x0AU,
    0x00U, 0x00U, 0x00U, 0x00U,
    0x28U,
    0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U
};

/**
  * @brief Sense data for 'No Error' (Used by reply data of Request Sense(03h) command)
  */
static USBD_MSC_SCSI_RET_SENSE_DATA_t l_tSense_Data_NoError = {
    0x00U,
    0x00U,
    0x00U,  /* No Sense */
    0x00U, 0x00U, 0x00U, 0x00U,
    0x00U,
    0x00U, 0x00U, 0x00U, 0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U,
    0x00U
};

/**
  * @brief Inquiry Data (Used by reply data of Inquiry(12h) command)
  */
static USBD_MSC_SCSI_RET_INQUIRY_DATA_t l_tInquiry_Data = {
    0x00U,
    0x80U,
    0x02U,
    0x02U,
    0x1FU,
    0x00U, 0x00U, 0x00U,
    0x54U, 0x4FU, 0x53U, 0x48U, 0x49U, 0x42U, 0x41U, 0x20U,                                                         /* TOSHIBA  */
    0x52U, 0x41U, 0x4DU, 0x2DU, 0x44U, 0x69U, 0x73U, 0x6BU, 0x20U, 0x20U, 0x20U, 0x20U, 0x20U, 0x20U, 0x20U, 0x20U, /* RAM-Disk */
    0x31U, 0x2EU, 0x30U, 0x30U                                                                                      /* 1.00     */
};

/**
  * @brief Read Capacity Data (Used by reply data of Read Capacity(25h) command)
  */
static USBD_MSC_SCSI_RET_READ_CAPACITY_DATA_t l_tReadCapacity_Data;

/**
  * @brief Mode Sense Data (Used by reply data of Mode Sense(6)(1Ah) command)
  */
static USBD_MSC_SCSI_RET_MODE_SENSE_DATA_t l_tSenseData;

/**
  * @brief Mode Sense Header (Used by Mode Sense data)
  */
static USBD_MSC_SCSI_RET_MODE_SENSE_HEADER_t l_tSenseData_Header = {
    0x03U,
    0x00U,
    0x00U,
    0x00U
};

/* Prototypes */
static void                                    _USBD_MSC_SCSI_CB_RecvCBW(uint8_t*);
static void                                    _USBD_MSC_SCSI_CB_SendData(uint32_t, uint8_t*, void*, uint16_t);
static void                                    _USBD_MSC_SCSI_CB_RecvData(uint32_t, uint8_t*, void*, uint16_t);
static void                                    _USBD_MSC_SCSI_Command_Analysis(uint8_t*);
static void                                    _USBD_MSC_SCSI_TestUnitReady(uint32_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_RequestSense(uint32_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_Inquiry(uint32_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_ModeSense(uint32_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_StartStopUnit(uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_PreventAllowMediumRemoval(uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_ReadCapacity(uint32_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_Read(uint32_t, uint8_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_Write(uint32_t, uint8_t, uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_Verify(uint8_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_NonSupported_Command(void);
static void                                    _USBD_MSC_SCSI_Res_RequestSense(uint32_t);
static void                                    _USBD_MSC_SCSI_Res_Inquiry(uint32_t);
static void                                    _USBD_MSC_SCSI_Res_ModeSense(uint32_t);
static void                                    _USBD_MSC_SCSI_Res_ReadCapacity(uint32_t);
static void                                    _USBD_MSC_SCSI_Res_Read(uint32_t, uint8_t*);
static void                                    _USBD_MSC_SCSI_Res_Write(uint32_t, uint8_t*);
static USBD_MSC_SCSI_RET_SENSE_DATA_t*         _USBD_MSC_SCSI_Get_Sense_Data(void);
static USBD_MSC_SCSI_RET_INQUIRY_DATA_t*       _USBD_MSC_SCSI_Get_Inquiry_Data(void);
static USBD_MSC_SCSI_RET_READ_CAPACITY_DATA_t* _USBD_MSC_SCSI_Get_ReadCapacity_Data(void);
static USBD_MSC_SCSI_RET_MODE_SENSE_DATA_t*    _USBD_MSC_SCSI_Get_ModeSense_Data(uint8_t, uint8_t);


/******************************************************************************/

/**
  * @brief  Initialize MSC(SCSI) driver
  * @param  p_scsi_cb : MSC(SCSI) callback information pointer.
  * @retval None
  */
void USBD_MSC_SCSI_Init(SCSI_CB_t *p_scsi_cb)
{
    /* Initialization of an internal variable */
    l_ucMediaCondition  = false;
    l_ucMediaState      = USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA;
    l_ucCSWStatus       = USBD_MSC_SCSI_COMMAND_PASSED;
    l_ucOperationCode   = 0U;
    l_ulDataLength      = 0U;
    l_ucSenseKey        = 0U;

    /* Set of callback functions */
    /* From MSC(SCSI) to application */
    gSCSI_CB.pfnMediaRead  = p_scsi_cb->pfnMediaRead;
    gSCSI_CB.pfnMediaWrite = p_scsi_cb->pfnMediaWrite;
    gSCSI_CB.pfnSendData   = p_scsi_cb->pfnSendData;
    gSCSI_CB.pfnRecvData   = p_scsi_cb->pfnRecvData;
    /* From MSC(BOT) to application */
    gBOT_CB.pfnConnStat         = p_scsi_cb->pfnConnStat;
    gBOT_CB.pfnMassStorageReset = p_scsi_cb->pfnMassStorageReset;
    gBOT_CB.pfnGetMaxLun        = p_scsi_cb->pfnGetMaxLun;
    gBOT_CB.pfnRecvCbw          = _USBD_MSC_SCSI_CB_RecvCBW;
    gBOT_CB.pfnSendData         = _USBD_MSC_SCSI_CB_SendData;
    gBOT_CB.pfnRecvData         = _USBD_MSC_SCSI_CB_RecvData;
    gBOT_CB.pfnResume           = NULL;
    gBOT_CB.pfnSuspend          = NULL;

    /* Initialize MSC(BOT) driver */
    usbd_msc_init(&gBOT_CB);

    return;
}

/**
  * @brief  Send read media data
  * @param  ulSize : Read media data size.
  * @param  pucData : Read media data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @retval bStat : true or false.
  */
bool USBD_MSC_SCSI_MediaReadData(uint32_t ulSize, uint8_t *pucData, void *pvInfo)
{
    bool bStat;

    if (gSCSI_CB.pfnSendData == NULL) {
        /* Callback function not found */
        return false;
    }

    /* Send data request */
    bStat = USBD_MSC_SendData(ulSize, pucData, pvInfo);

    return bStat;
}

/**
  * @brief  Receive write media data
  * @param  ulSize : Write media data size.
  * @param  pucData : Write media data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @retval bStat : true or false.
  */
bool USBD_MSC_SCSI_MediaWriteData(uint32_t ulSize, uint8_t *pucData, void *pvInfo)
{
    bool bStat;

    if (gSCSI_CB.pfnRecvData == NULL) {
        /* Callback function not found */
        return false;
    }

    /* Receive data request */
    bStat = USBD_MSC_RecvData(ulSize, pucData, pvInfo);

    return bStat;
}

/**
  * @brief  Send result for read media data
  * @param  ucResult : Processed result.
  *    This parameter can be one of the following values :
  *      USBD_MSC_SCSI_COMMAND_PASSED,
  *      USBD_MSC_SCSI_COMMAND_FAILED,
  *      USBD_MSC_SCSI_COMMAND_PHASE_ERROR
  * @param  ulSize : Processed data size.
  * @param  bStall : STALL flag.
  *    This parameter can be one of the following values :
  *      USBD_MSC_SCSI_REQ_STALL,
  *      USBD_MSC_SCSI_NO_STALL
  * @retval None
  */
void USBD_MSC_SCSI_MediaReadResult(uint8_t ucResult, uint32_t ulSize, bool bStall)
{
    /* Reply status */
    USBD_MSC_SetResponse(ucResult, ulSize, bStall);
}

/**
  * @brief  Send result for write media data
  * @param  ucResult : Processed result.
  *    This parameter can be one of the following values :
  *      USBD_MSC_SCSI_COMMAND_PASSED,
  *      USBD_MSC_SCSI_COMMAND_FAILED,
  *      USBD_MSC_SCSI_COMMAND_PHASE_ERROR
  * @param  ulSize : Processed data size.
  * @param  bStall : STALL flag.
  *    This parameter can be one of the following values :
  *      USBD_MSC_SCSI_REQ_STALL,
  *      USBD_MSC_SCSI_NO_STALL
  * @retval None
  */
void USBD_MSC_SCSI_MediaWriteResult(uint8_t ucResult, uint32_t ulSize, bool bStall)
{
    /* Reply status */
    USBD_MSC_SetResponse(ucResult, ulSize, bStall);
}

/**
  * @brief  Set media connect/disconnect status
  * @param  ucConnect : Media condition.
  * @retval None
  */
void USBD_MSC_SCSI_Media_ConnStat(uint8_t ucConnect)
{
    /* Set media condition */
    l_ucMediaCondition = ucConnect;

    if (ucConnect == true) {
        l_ucMediaState = USBD_MSC_SCSI_MEDIASTATE_OK;
        /* Set Sense Key : Unit Attention(06h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_UNIT_ATTENTION;
    } else if (ucConnect == false) {
        l_ucMediaState = USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA;
    }
}

/**
  * @brief  Set sense key
  * @param  ucSenseKey : Sense key.
  * @retval None
  */
void USBD_MSC_SCSI_SetSenseKey(uint8_t ucSenseKey)
{
    /* Set sense key */
    l_ucSenseKey = ucSenseKey;
}

/**
  * @brief  Callback for received CBW
  * @param  pucData : CBW buffer pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_CB_RecvCBW(uint8_t *pucData)
{
    /* SCSI command analysis */
    _USBD_MSC_SCSI_Command_Analysis(pucData);
}

/**
  * @brief  Callback for Send Data
  * @param  ulSize : Send data size.
  * @param  pucData : Send data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @param  usStat : Send status.
  *   This parameter can be one of the following values :
  *      EPSTAT_COMPLETE, EPSTAT_ABORT
  * @retval None
  */
static void _USBD_MSC_SCSI_CB_SendData(uint32_t ulSize, uint8_t *pucData, void *pvInfo, uint16_t usStat)
{
    if (usStat == EPSTAT_COMPLETE) {
        if ((uint32_t)l_ucOperationCode != (uint32_t)pvInfo) {
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            return;
        }

        switch((uint32_t)pvInfo) {
        case USBD_MSC_SCSI_REQUEST_SENCE:
            /* Request Sense(03h) */
            _USBD_MSC_SCSI_Res_RequestSense(ulSize);
            break;
        case USBD_MSC_SCSI_INQUIRY:
            /* Inquiry(12h) */
            _USBD_MSC_SCSI_Res_Inquiry(ulSize);
            break;
        case USBD_MSC_SCSI_MODE_SENSE:
            /* Mode Sense(6)(1Ah) */
            _USBD_MSC_SCSI_Res_ModeSense(ulSize);
            break;
        case USBD_MSC_SCSI_READ_CAPACITY:
            /* Read Capacity(25h) */
            _USBD_MSC_SCSI_Res_ReadCapacity(ulSize);
            break;
        case USBD_MSC_SCSI_READ:
            /* Read(10)(28h) */
            _USBD_MSC_SCSI_Res_Read(ulSize, pucData);
            break;
        default:
            /* Others */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            break;
        }
    }
}

/**
  * @brief  Callback for received Data
  * @param  ulSize : Received data size.
  * @param  pucData : Received data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @param  usStat : Send status.
  *   This parameter can be one of the following values :
  *      EPSTAT_COMPLETE, EPSTAT_ABORT
  * @retval None
  */
static void _USBD_MSC_SCSI_CB_RecvData(uint32_t ulSize, uint8_t *pucData, void *pvInfo, uint16_t usStat)
{
    if (usStat == EPSTAT_COMPLETE) {
        if ((uint32_t)l_ucOperationCode != (uint32_t)pvInfo) {
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            return;
        }

        switch((uint32_t)pvInfo) {
        case USBD_MSC_SCSI_WRITE:
            /* Write(10)(2Ah) */
            _USBD_MSC_SCSI_Res_Write(ulSize, pucData);
            break;
        default:
            /* Others */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            break;
        }
    }
}

/**
  * @brief  SCSI command Analysis
  * @param  pucData : CBW buffer pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Command_Analysis(uint8_t *pucData)
{
    /* CBWCB */
    CBW_t *ptCBW = (CBW_t *)pucData;

    l_ulDataLength    = 0U;
    l_ucOperationCode = 0U;

    /* Check LUN */
    if (ptCBW->ucCBWLUN != 0U) {
        /* LUN is only 0 in sample */
        _USBD_MSC_SCSI_NonSupported_Command();
        return;
    }

    l_ulDataLength = ptCBW->ulCBWDataTransferLength;
    l_ucOperationCode = ptCBW->aucCBWCB[0U];

    switch(ptCBW->aucCBWCB[0U]) {
    case USBD_MSC_SCSI_TEST_UNIT_READY:
        /* Test Unit Ready(00h) */
        _USBD_MSC_SCSI_TestUnitReady(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_REQUEST_SENCE:
        /* Request Sense(03h) */
        _USBD_MSC_SCSI_RequestSense(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_INQUIRY:
        /* Inquiry(12h) */
        _USBD_MSC_SCSI_Inquiry(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_MODE_SENSE:
        /* Mode Sense(6)(1Ah) */
        _USBD_MSC_SCSI_ModeSense(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_START_STOP_UNIT:
        /* Start Stop Unit(1Bh) */
        _USBD_MSC_SCSI_StartStopUnit(ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL:
        /* Prevent/Allow Medium Removal(1Eh) */
        _USBD_MSC_SCSI_PreventAllowMediumRemoval(ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_READ_CAPACITY:
        /* Read Capacity(25h) */
        _USBD_MSC_SCSI_ReadCapacity(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_READ:
        /* Read(10)(28h) */
        _USBD_MSC_SCSI_Read(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWFlag, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_WRITE:
        /* Write(10)(2Ah) */
        _USBD_MSC_SCSI_Write(ptCBW->ulCBWDataTransferLength, ptCBW->ucCBWFlag, ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    case USBD_MSC_SCSI_VERIFY:
        /* Verify(2Fh) */
        _USBD_MSC_SCSI_Verify(ptCBW->ucCBWCBLength, ptCBW->aucCBWCB);
        break;
    default:
        /* Others */
        _USBD_MSC_SCSI_NonSupported_Command();
        break;
    }
}

/**
  * @brief  Test Unit Ready(00h) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_TestUnitReady(uint32_t ulTransferDataLength, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    if ( (l_ucMediaCondition == false)
     && (l_ucMediaState == USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA)) {
        /* Set Sense Key : Not Ready(02h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENCE_KEY_NOT_READY;
        /* Reply status */
        USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
    } else {
        if (ulTransferDataLength == 0) {
            /* Reply status */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PASSED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
        } else {
            /* Set Sense Key : Illegal Request(05h) */
            l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST;
            /* Reply status (FAILED) */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        }
    }
}

/**
  * @brief  Request Sense(03h) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_RequestSense(uint32_t ulTransferDataLength, uint8_t ucCBWCBLength, uint8_t *pucCBWCB )
{
    USBD_MSC_SCSI_RET_SENSE_DATA_t *ptSense;
    uint8_t                        ucAllocSize;

    /* Get sense data */
    ptSense = _USBD_MSC_SCSI_Get_Sense_Data();
    ucAllocSize = *(pucCBWCB + 4U);

    /* Reply sense data */
    gbStatDCSCSI = USBD_MSC_SendData((uint32_t)ucAllocSize, (uint8_t *)ptSense, (void *)l_ucOperationCode);
    /* Don't care status */
}

/**
  * @brief  Inquiry(12h) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Inquiry(uint32_t ulTransferDataLength, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    USBD_MSC_SCSI_RET_INQUIRY_DATA_t *ptInquiry;
    uint8_t                          ucAllocSize;

    /* Get inquiry data */
    ptInquiry = _USBD_MSC_SCSI_Get_Inquiry_Data();
    ucAllocSize = *(pucCBWCB + 4U);

    /* Set Sense Key : No Sense(00h) */
    l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_NO_SENSE;

    /* Reply inquiry data */
    gbStatDCSCSI = USBD_MSC_SendData((uint32_t)ucAllocSize, (uint8_t *)ptInquiry, (void *)l_ucOperationCode);
    /* Don't care status */
}

/**
  * @brief  Mode Sense(6)(1Ah) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_ModeSense(uint32_t ulTransferDataLength, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    USBD_MSC_SCSI_RET_MODE_SENSE_DATA_t *ptModeSense;
    uint32_t                            ulSize;
    uint8_t                             ucPC;
    uint8_t                             ucPageCode;

    if ((l_ucMediaCondition == false) && (l_ucMediaState == USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA)) {
        /* Set Sense Key ad Not Ready(02h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENCE_KEY_NOT_READY;
        /* Reply status */
        USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        return;

    } else {
        /* PC */
        ucPC = (uint8_t)(*(pucCBWCB + 2U) & USBD_MSC_SCSI_PAGE_CONTROL_BIT);
        /* Page code */
        ucPageCode = (uint8_t)(*(pucCBWCB + 2U) & USBD_MSC_SCSI_PAGE_CODE_BIT);

        /* Get mode sense data */
        ptModeSense = _USBD_MSC_SCSI_Get_ModeSense_Data(ucPC, ucPageCode);
        ulSize = sizeof(USBD_MSC_SCSI_RET_MODE_SENSE_HEADER_t);
        
        /* Reply mode sense data */
        gbStatDCSCSI = USBD_MSC_SendData(ulSize, (uint8_t *)ptModeSense, (void *)l_ucOperationCode);
        /* Don't care status */
    }
}

/**
  * @brief  Start Stop Unit(1Bh) command process
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_StartStopUnit(uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    uint8_t ucLoEj;
    uint8_t ucStart;

    /* Get LoEj bit */
    ucLoEj = (uint8_t)((*(pucCBWCB + 4U) & USBD_MSC_SCSI_LOEJ_BIT) >> 1U);
    /* Get Start bit */
    ucStart = (uint8_t)(*(pucCBWCB + 4U) & USBD_MSC_SCSI_START_BIT);

    /* Check LoEj and Start bit */
    if ((ucLoEj == USBD_MSC_SCSI_AUTO_LOAD_EJECT_START) && (ucStart == USBD_MSC_SCSI_MEDIA_ACCESS_STOP)) {
        /* Check Prevent bit */
        if (l_ucPrevent == USBD_MSC_SCSI_MEDIA_REMOVAL_OK) {
            /* Set disconnect media status */
            USBD_MSC_SCSI_Media_ConnStat(false);
        } else {
            /* Set Sense Key : Unit Attention(06h) */
            l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_UNIT_ATTENTION;
            /* Reply status */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
            return;

        }
    } else if ((ucLoEj == USBD_MSC_SCSI_AUTO_LOAD_EJECT_START) && (ucStart == USBD_MSC_SCSI_MEDIA_ACCESS_START)) {
        /* Set connect media status */
        USBD_MSC_SCSI_Media_ConnStat(true);
    }

    /* Reply status */
    USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PASSED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
}

/**
  * @brief  Prevent/Allow Medium Removal(1Eh) command process
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_PreventAllowMediumRemoval(uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    /* Get Prevent bit */
    l_ucPrevent = (uint8_t)(*(pucCBWCB + 4U) & USBD_MSC_SCSI_PREVENT_BIT);

    /* Reply status */
    USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PASSED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
}

/**
  * @brief  Read Capacity(25h) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_ReadCapacity(uint32_t ulTransferDataLength, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    USBD_MSC_SCSI_RET_READ_CAPACITY_DATA_t *ptReadCapacity;
    uint32_t                               ulSize;
    uint32_t                               ulLBA;
    uint8_t                                ucPMI;

    if ((l_ucMediaCondition == false) && (l_ucMediaState == USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA)) {
        /* Set Sense Key : Not Ready(02h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENCE_KEY_NOT_READY;
        /* Reply status */
        USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        return;

    } else {
        /* PMI */
        ucPMI = (uint8_t)(*(pucCBWCB + 8U) & USBD_MSC_SCSI_PMI_BIT);
        /* LBA */
        ulLBA = *(pucCBWCB + 2U);
        ulLBA <<= 8U;
        ulLBA |= *(pucCBWCB + 3U);
        ulLBA <<= 8U;
        ulLBA |= *(pucCBWCB + 4U);
        ulLBA <<= 8U;
        ulLBA |= *(pucCBWCB + 5U);

        if (ucPMI == 0U && ulLBA != 0U) {
            /* Set Sense Key : Illegal Request(05h) */
            l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST;
            /* Reply status */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            return;
        }

        /* Get read capacity data */
        ptReadCapacity = _USBD_MSC_SCSI_Get_ReadCapacity_Data();
        ulSize = sizeof(USBD_MSC_SCSI_RET_READ_CAPACITY_DATA_t);

        /* Reply read capacity data */
        gbStatDCSCSI = USBD_MSC_SendData(ulSize, (uint8_t *)ptReadCapacity, (void *)l_ucOperationCode);
        /* Don't care status */
    }
}

/**
  * @brief  Read(10)(28h) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucDir : Transfer direction.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Read(uint32_t ulTransferDataLength, uint8_t ucDir, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    uint32_t ulLBA;
    uint16_t usBlockNum;

    if ((l_ucMediaCondition == false) && (l_ucMediaState == USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA)) {
        /* Set Sense Key : Not Ready(02h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENCE_KEY_NOT_READY;
        /* Reply status */
        USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        return;

    } else {
        if ((ucDir & USBD_MSC_BOT_CBW_FLAGS_DBIT) != USBD_MSC_BOT_CBW_FLAGS_D2H) {
            /* Reply status (PHASE_ERROR) */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        } else {
            /* Get LBA */
            ulLBA = *(pucCBWCB + 2U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 3U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 4U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 5U);
            /* Get Blocks */
            usBlockNum = *(pucCBWCB + 7U);
            usBlockNum <<= 8U;
            usBlockNum |= *(pucCBWCB + 8U);

            if (ulTransferDataLength == 0) {
                /* Reply status (PHASE_ERROR) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else if (ulTransferDataLength > (usBlockNum * USBD_MSC_SCSI_BLOCK_SIZE)) {
                /* Set Sense Key : Illegal Request(05h) */
                l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST;
                /* Reply status (FAILED) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else if (ulTransferDataLength < (usBlockNum * USBD_MSC_SCSI_BLOCK_SIZE)) {
                /* Reply status (PHASE_ERROR) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else {
                if (gSCSI_CB.pfnMediaRead != NULL) {
                    /* Callback */
                    (*gSCSI_CB.pfnMediaRead)(ulTransferDataLength, ulLBA, usBlockNum);
                }
            }
        }
    }
}

/**
  * @brief  Write(10)(2Ah) command process
  * @param  ulTransferDataLength : Transfer data length.
  * @param  ucDir : Transfer direction.
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Write(uint32_t ulTransferDataLength, uint8_t ucDir, uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    uint32_t ulLBA;
    uint16_t usBlockNum;

    if ((l_ucMediaCondition == false) && (l_ucMediaState == USBD_MSC_SCSI_MEDIASTATE_NO_MEDIA)) {
        /* Set Sense Key : Not Ready(02h) */
        l_ucSenseKey = USBD_MSC_SCSI_SENCE_KEY_NOT_READY;
        /* Reply status */
        USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        return;

    } else {
        if ((ucDir & USBD_MSC_BOT_CBW_FLAGS_DBIT) == USBD_MSC_BOT_CBW_FLAGS_D2H) {
            /* Reply status (PHASE_ERROR) */
            USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
        } else {
            /* Get LBA */
            ulLBA = *(pucCBWCB + 2U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 3U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 4U);
            ulLBA <<= 8U;
            ulLBA |= *(pucCBWCB + 5U);
            /* Get Blocks */
            usBlockNum = *(pucCBWCB + 7U);
            usBlockNum <<= 8U;
            usBlockNum |= *(pucCBWCB + 8U);

            if (ulTransferDataLength == 0) {
                /* Reply status (PHASE_ERROR) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else if (ulTransferDataLength > (usBlockNum * USBD_MSC_SCSI_BLOCK_SIZE)) {
                /* Set Sense Key : Illegal Request(05h) */
                l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST;
                /* Reply status (FAILED) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else if (ulTransferDataLength < (usBlockNum * USBD_MSC_SCSI_BLOCK_SIZE)) {
                /* Reply status (PHASE_ERROR) */
                USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PHASE_ERROR, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
            } else {
                if (gSCSI_CB.pfnMediaWrite != NULL) {
                    /* Callback */
                    (*gSCSI_CB.pfnMediaWrite)(ulTransferDataLength, ulLBA, usBlockNum);
                }
            }
        }
    }
}

/**
  * @brief  Verify(2Fh) command process
  * @param  ucCBWCBLength : CBW command block length.
  * @param  pucCBWCB : CBW command block pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Verify(uint8_t ucCBWCBLength, uint8_t *pucCBWCB)
{
    /* Force pass in sample */
    /* Reply status */
    USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PASSED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_NO_STALL);
}

/**
  * @brief  Non supported command process
  * @param  None
  * @retval None
  */
static void _USBD_MSC_SCSI_NonSupported_Command(void)
{
    /* Set Sense Key : Illegal Request(05h) */
    l_ucSenseKey = USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST;

    /* Reply status */
    USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_FAILED, USBD_MSC_SCSI_NO_DATA, USBD_MSC_SCSI_REQ_STALL);
}

/**
  * @brief  Request Sense(03h) command status process
  * @param  ulSize : Request data size.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_RequestSense(uint32_t ulSize)
{
    bool bStall = USBD_MSC_SCSI_NO_STALL;

    /* Compare with send size and request size */
    if (ulSize == l_ulDataLength) {
        /* Status is COMMAND_PASSED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PASSED;
        /* No STALL */
        bStall = USBD_MSC_SCSI_NO_STALL;
    } else if (ulSize < l_ulDataLength) {
        /* Status is COMMAND_FAILED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_FAILED;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    } else if (ulSize > l_ulDataLength) {
        /* Status is PHASE_ERROR */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PHASE_ERROR;
        /* Clear data size */
        ulSize = USBD_MSC_SCSI_NO_DATA;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    }

    /* Reply status */
    USBD_MSC_SetResponse(l_ucCSWStatus, ulSize, bStall);
}

/**
  * @brief  Inquiry(12h) command status process
  * @param  ulSize : Request data size.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_Inquiry(uint32_t ulSize)
{
    bool bStall = USBD_MSC_SCSI_NO_STALL;

    /* Compare with send size and request size */
    if (ulSize == l_ulDataLength) {
        /* Status is COMMAND_PASSED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PASSED;
        /* No STALL */
        bStall = USBD_MSC_SCSI_NO_STALL;
    } else if (ulSize < l_ulDataLength) {
        /* Status is COMMAND_FAILED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_FAILED;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    } else if (ulSize > l_ulDataLength) {
        /* Status is PHASE_ERROR */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PHASE_ERROR;
        /* Clear data size */
        ulSize = USBD_MSC_SCSI_NO_DATA;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    }

    /* Reply status */
    USBD_MSC_SetResponse(l_ucCSWStatus, ulSize, bStall);
}

/**
  * @brief  Mode Sense(6)(1Ah) command status process
  * @param  ulSize : Request data size.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_ModeSense(uint32_t ulSize)
{
    /* Reply status */
    USBD_MSC_SetResponse(USBD_MSC_SCSI_COMMAND_PASSED, ulSize, USBD_MSC_SCSI_NO_STALL);
}

/**
  * @brief  Read Capacity(25h) command status process
  * @param  ulSize : Request data size.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_ReadCapacity(uint32_t ulSize)
{
    bool    bStall = USBD_MSC_SCSI_NO_STALL;

    /* Compare with send size and request size */
    if (ulSize == l_ulDataLength) {
        /* Status is COMMAND_PASSED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PASSED;
        /* No STALL */
        bStall = USBD_MSC_SCSI_NO_STALL;
    } else if (ulSize < l_ulDataLength) {
        /* Status is COMMAND_FAILED */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_FAILED;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    } else if (ulSize > l_ulDataLength) {
        /* Status is PHASE_ERROR */
        l_ucCSWStatus = USBD_MSC_SCSI_COMMAND_PHASE_ERROR;
        /* Clear data size */
        ulSize = USBD_MSC_SCSI_NO_DATA;
        /* Set STALL */
        bStall = USBD_MSC_SCSI_REQ_STALL;
    }

    /* Reply status */
    USBD_MSC_SetResponse(l_ucCSWStatus, ulSize, bStall);
}

/**
  * @brief  Read(10)(28h) command status process
  * @param  ulSize : Request data size.
  * @param  pucData : Request data buffer pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_Read(uint32_t ulSize, uint8_t *pucData)
{
    if (gSCSI_CB.pfnSendData != NULL) {
        /* Callback */
        (*gSCSI_CB.pfnSendData)(ulSize, pucData);
    }
}

/**
  * @brief  Write(10)(2Ah) command status process
  * @param  ulSize : Request data size.
  * @param  pucData : Request data buffer pointer.
  * @retval None
  */
static void _USBD_MSC_SCSI_Res_Write(uint32_t ulSize, uint8_t *pucData)
{
    if (gSCSI_CB.pfnRecvData != NULL) {
        /* Callback */
        (*gSCSI_CB.pfnRecvData)(ulSize, pucData);
    }
}

/**
  * @brief  Get sense data
  * @param  None
  * @retval ptSense_Data : Sense data pointer.
  */
static USBD_MSC_SCSI_RET_SENSE_DATA_t* _USBD_MSC_SCSI_Get_Sense_Data(void)
{
    USBD_MSC_SCSI_RET_SENSE_DATA_t *ptSense_Data;

    /* Fixed data in sample */
    switch(l_ucSenseKey) {
    case USBD_MSC_SCSI_SENCE_KEY_NOT_READY:
        /* Not Ready(02h) */
        ptSense_Data = &l_tSense_Data_NoMedia;
        break;
    case USBD_MSC_SCSI_SENSE_KEY_ILLEGAL_REQUEST:
        /* Illegal Request(05h) */
        ptSense_Data = &l_tSense_Data_Error;
        break;
    case USBD_MSC_SCSI_SENSE_KEY_UNIT_ATTENTION:
        /* Unit Attention(06h) */
        ptSense_Data = &l_tSense_Data_MediaChenge;
        break;
    default:
        /* No Sense(00h) */
        ptSense_Data = &l_tSense_Data_NoError;
        break;
    }

    return ptSense_Data;
}

/**
  * @brief  Get inquiry data
  * @param  None
  * @retval l_tInquiry_Data : Inquiry data pointer.
  */
static USBD_MSC_SCSI_RET_INQUIRY_DATA_t* _USBD_MSC_SCSI_Get_Inquiry_Data(void)
{
    /* Fixed data in sample */
    return &l_tInquiry_Data;
}

/**
  * @brief  Get read capacity data
  * @param  None
  * @retval l_tReadCapacity_Data : Read capacity data pointer.
  */
static USBD_MSC_SCSI_RET_READ_CAPACITY_DATA_t* _USBD_MSC_SCSI_Get_ReadCapacity_Data(void)
{
    /* Last LBA */
    l_tReadCapacity_Data.uc_LogicalBlockAddr[0] = (uint8_t)((USBD_MSC_SCSI_BLOCK_NUM - 1) >> 24U);
    l_tReadCapacity_Data.uc_LogicalBlockAddr[1] = (uint8_t)((USBD_MSC_SCSI_BLOCK_NUM - 1) >> 16U);
    l_tReadCapacity_Data.uc_LogicalBlockAddr[2] = (uint8_t)((USBD_MSC_SCSI_BLOCK_NUM - 1) >>  8U);
    l_tReadCapacity_Data.uc_LogicalBlockAddr[3] = (uint8_t)(USBD_MSC_SCSI_BLOCK_NUM - 1);
    /* Block length */
    l_tReadCapacity_Data.uc_BlockLength[0] = (uint8_t)(USBD_MSC_SCSI_BLOCK_SIZE >> 24U);
    l_tReadCapacity_Data.uc_BlockLength[1] = (uint8_t)(USBD_MSC_SCSI_BLOCK_SIZE >> 16U);
    l_tReadCapacity_Data.uc_BlockLength[2] = (uint8_t)(USBD_MSC_SCSI_BLOCK_SIZE >>  8U);
    l_tReadCapacity_Data.uc_BlockLength[3] = (uint8_t)USBD_MSC_SCSI_BLOCK_SIZE;
    
    return &l_tReadCapacity_Data;
}

/**
  * @brief  Get mode sense data
  * @param  None
  * @retval l_tSenseData : Sense data pointer.
  */
static USBD_MSC_SCSI_RET_MODE_SENSE_DATA_t* _USBD_MSC_SCSI_Get_ModeSense_Data(uint8_t ucPC, uint8_t ucPageCode)
{
    /* Fixed data in sample */
    l_tSenseData.tHeader = l_tSenseData_Header;

    return &l_tSenseData;
}

/* EOC */
